#pragma once

// @generated by torchgen/gen.py from Function.h

#include <ATen/Context.h>
#include <ATen/DeviceGuard.h>
#include <ATen/TensorUtils.h>
#include <ATen/TracerMode.h>
#include <ATen/core/Generator.h>
#include <ATen/core/Reduction.h>
#include <ATen/core/Tensor.h>
#include <c10/core/Scalar.h>
#include <c10/core/Storage.h>
#include <c10/core/TensorOptions.h>
#include <c10/util/Deprecated.h>
#include <c10/util/Optional.h>



#include <ATen/ops/set_ops.h>

namespace at {


// aten::set.source_Storage_out(Tensor self, Storage source, *, Tensor(a!) out) -> Tensor(a!)
inline at::Tensor & set_out(at::Tensor & out, const at::Tensor & self, at::Storage source) {
    return at::_ops::set_source_Storage_out::call(self, source, out);
}

// aten::set.source_Storage_out(Tensor self, Storage source, *, Tensor(a!) out) -> Tensor(a!)
inline at::Tensor & set_outf(const at::Tensor & self, at::Storage source, at::Tensor & out) {
    return at::_ops::set_source_Storage_out::call(self, source, out);
}

// aten::set.source_Storage(Tensor self, Storage source) -> Tensor
inline at::Tensor set(const at::Tensor & self, at::Storage source) {
    return at::_ops::set_source_Storage::call(self, source);
}

// aten::set.source_Storage_storage_offset_out(Tensor self, Storage source, SymInt storage_offset, SymInt[] size, SymInt[] stride=[], *, Tensor(a!) out) -> Tensor(a!)
inline at::Tensor & set_out(at::Tensor & out, const at::Tensor & self, at::Storage source, int64_t storage_offset, at::IntArrayRef size, at::IntArrayRef stride={}) {
    return at::_ops::set_source_Storage_storage_offset_out::call(self, source, storage_offset, c10::fromIntArrayRef(size), c10::fromIntArrayRef(stride), out);
}

// aten::set.source_Storage_storage_offset_out(Tensor self, Storage source, SymInt storage_offset, SymInt[] size, SymInt[] stride=[], *, Tensor(a!) out) -> Tensor(a!)
inline at::Tensor & set_outf(const at::Tensor & self, at::Storage source, int64_t storage_offset, at::IntArrayRef size, at::IntArrayRef stride, at::Tensor & out) {
    return at::_ops::set_source_Storage_storage_offset_out::call(self, source, storage_offset, c10::fromIntArrayRef(size), c10::fromIntArrayRef(stride), out);
}

// aten::set.source_Storage_storage_offset_out(Tensor self, Storage source, SymInt storage_offset, SymInt[] size, SymInt[] stride=[], *, Tensor(a!) out) -> Tensor(a!)
inline at::Tensor & set_symint_out(at::Tensor & out, const at::Tensor & self, at::Storage source, c10::SymInt storage_offset, c10::SymIntArrayRef size, c10::SymIntArrayRef stride={}) {
    return at::_ops::set_source_Storage_storage_offset_out::call(self, source, storage_offset, size, stride, out);
}

// aten::set.source_Storage_storage_offset_out(Tensor self, Storage source, SymInt storage_offset, SymInt[] size, SymInt[] stride=[], *, Tensor(a!) out) -> Tensor(a!)
inline at::Tensor & set_symint_outf(const at::Tensor & self, at::Storage source, c10::SymInt storage_offset, c10::SymIntArrayRef size, c10::SymIntArrayRef stride, at::Tensor & out) {
    return at::_ops::set_source_Storage_storage_offset_out::call(self, source, storage_offset, size, stride, out);
}

// aten::set.source_Storage_storage_offset(Tensor self, Storage source, SymInt storage_offset, SymInt[] size, SymInt[] stride=[]) -> Tensor
inline at::Tensor set(const at::Tensor & self, at::Storage source, int64_t storage_offset, at::IntArrayRef size, at::IntArrayRef stride={}) {
    return at::_ops::set_source_Storage_storage_offset::call(self, source, storage_offset, c10::fromIntArrayRef(size), c10::fromIntArrayRef(stride));
}

// aten::set.source_Storage_storage_offset(Tensor self, Storage source, SymInt storage_offset, SymInt[] size, SymInt[] stride=[]) -> Tensor
inline at::Tensor set_symint(const at::Tensor & self, at::Storage source, c10::SymInt storage_offset, c10::SymIntArrayRef size, c10::SymIntArrayRef stride={}) {
    return at::_ops::set_source_Storage_storage_offset::call(self, source, storage_offset, size, stride);
}

// aten::set.source_Tensor_out(Tensor self, Tensor source, *, Tensor(a!) out) -> Tensor(a!)
inline at::Tensor & set_out(at::Tensor & out, const at::Tensor & self, const at::Tensor & source) {
    return at::_ops::set_source_Tensor_out::call(self, source, out);
}

// aten::set.source_Tensor_out(Tensor self, Tensor source, *, Tensor(a!) out) -> Tensor(a!)
inline at::Tensor & set_outf(const at::Tensor & self, const at::Tensor & source, at::Tensor & out) {
    return at::_ops::set_source_Tensor_out::call(self, source, out);
}

// aten::set.source_Tensor(Tensor self, Tensor source) -> Tensor
inline at::Tensor set(const at::Tensor & self, const at::Tensor & source) {
    return at::_ops::set_source_Tensor::call(self, source);
}

// aten::set.out(Tensor self, *, Tensor(a!) out) -> Tensor(a!)
inline at::Tensor & set_out(at::Tensor & out, const at::Tensor & self) {
    return at::_ops::set_out::call(self, out);
}

// aten::set.out(Tensor self, *, Tensor(a!) out) -> Tensor(a!)
inline at::Tensor & set_outf(const at::Tensor & self, at::Tensor & out) {
    return at::_ops::set_out::call(self, out);
}

// aten::set(Tensor self) -> Tensor
inline at::Tensor set(const at::Tensor & self) {
    return at::_ops::set::call(self);
}

}
